__title__ = 'magna'
__description__ = 'Magna is a collection of commands that I frequently use in bioinformatics and day-to-day life.'
__url__ = 'https://git.microba.core/modules/python/bio-proc-sdk'
__version__ = '2.7.0'
__author__ = 'Aaron Mussig'
__author_email__ = 'aaronmussig@gmail.com'
__license__ = 'GNU AGPLv3'
__bug_url__ = 'https://github.com/aaronmussig/magna'
__doc_url__ = 'https://github.com/aaronmussig/magna'
__src_url__ = 'https://github.com/aaronmussig/magna'



import numpy as np
import pandas as pd


def optimise_df(df: pd.DataFrame, integers: bool = True, floats: bool = True):
    """Optimise a Pandas DataFrame by using the smallest possible data type.

    Args:
        df: The Pandas DataFrame to optimise.
        integers: Whether to optimise the integers.
        floats: Whether to optimise the floats.
    """
    if floats:
        float_types = [np.float16, np.float32]
        float_cols = df.select_dtypes('float').columns
        for float_col in float_cols:
            for float_type in float_types:
                if np.min(df[float_col] - df[float_col].astype(float_type)) == 0:
                    df[float_col] = df[float_col].astype(float_type)
                    break

    if integers:
        sint_types = [np.int8, np.int16, np.int32, np.int64]
        uint_types = [np.uint8, np.uint16, np.uint32, np.uint64]
        sint_info = [np.iinfo(t) for t in sint_types]
        uint_info = [np.iinfo(t) for t in uint_types]

        int_cols = df.select_dtypes('integer').columns
        for int_col in int_cols:
            col_min, col_max = df[int_col].min(), df[int_col].max()

            # Determine which data types to use
            if col_min >= 0:
                int_info = uint_info
            else:
                int_info = sint_info

            # Set the data type
            for info in int_info:
                if col_min >= info.min and col_max <= info.max:
                    df[int_col] = df[int_col].astype(info.dtype)
                    break



from typing import Collection

import numpy as np
from scipy.cluster import hierarchy


def _get_newick(node, parent_dist, leaf_names, newick='') -> str:
    if node.is_leaf():
        return "%s:%.2f%s" % (leaf_names[node.id], parent_dist - node.dist, newick)
    else:
        if len(newick) > 0:
            newick = "):%.2f%s" % (parent_dist - node.dist, newick)
        else:
            newick = ");"
        newick = _get_newick(node.get_left(), node.dist, leaf_names, newick=newick)
        newick = _get_newick(node.get_right(), node.dist, leaf_names, newick=",%s" % newick)
        newick = "(%s" % newick
        return newick


def dm_to_newick(arr: np.ndarray, labels: Collection[str]) -> str:
    """Convert a pairwise distance matrix into Newick format.

    Args:
        arr: The symmetrical pairwise distance matrix.
        labels: The labels of the leaf nodes.
    """
    Z = hierarchy.linkage(arr)
    tree = hierarchy.to_tree(Z, rd=False)
    newick = _get_newick(tree, tree.dist, labels)
    return newick



import hashlib
import os
import shutil
import tarfile
import urllib
import urllib.request
from typing import Optional

from tqdm import tqdm

from magna.config import CACHE_DIR


class TqdmUpTo(tqdm):
    """Provides `update_to(n)` which uses `tqdm.update(delta_n)`."""

    def update_to(self, b=1, bsize=1, tsize=None):
        """
        b  : int, optional
            Number of blocks transferred so far [default: 1].
        bsize  : int, optional
            Size of each block (in tqdm units) [default: 1].
        tsize  : int, optional
            Total size (in tqdm units). If [default: None] remains unchanged.
        """
        if tsize is not None:
            self.total = tsize
        return self.update(b * bsize - self.n)  # also sets self.n = b * bsize


def untar(file_path: str, dir_name: str):
    """Extracts the contents of the tar file into the target directory.

    Args:
        file_path: The path to the tar file.
        dir_name: The directory to extract the tar file into.

    Examples:
        Extract the contents of the tar file at ``/tmp/data.tar.gz`` into the
        directory ``/tmp/data``.

        >>> untar('/tmp/data.tar.gz', '/tmp/data')
    """
    os.makedirs(dir_name, exist_ok=True)
    with tarfile.open(file_path) as tar:
        tar.extractall(dir_name)


def md5sum(path: str) -> str:
    """Returns the md5 hash of a file.

    Args:
        path: The path to the file.
    """
    block_size = 65536
    hasher = hashlib.md5()
    with open(path, 'rb') as f:
        buf = f.read(block_size)
        while len(buf) > 0:
            hasher.update(buf)
            buf = f.read(block_size)
    return hasher.hexdigest()


def download_file(url: str, path: str, md5: Optional[str] = None):
    """Downloads a file to disk, optionally validating the md5 hash.

    Args:
        url: The url to download from.
        path: The path to save the file to.
        md5: The expected md5 hash of the file.

    Raises:
        IOError: If the md5 hash doesn't match.

    Examples:
        Download the file at ``https://www.example.com/data.csv`` to ``/tmp/data.csv``.

        >>> download_file('https://www.example.com/data.csv', '/tmp/data.csv')
    """
    with TqdmUpTo(unit='B', unit_scale=True, unit_divisor=1024, miniters=1, desc=url.split('/')[-1]) as t:
        urllib.request.urlretrieve(url, filename=path, reporthook=t.update_to, data=None)
        t.total = t.n

    if md5 and md5 != md5sum(path):
        raise IOError('Hash mismatch')


def cache_file(srv_path: str, local_name: str) -> str:
    """Copies a file to the magna cache (doesn't auto-remove).

    Args:
        srv_path: The remote path of the file.
        local_name: The key to cache this file with.

    Returns:
        The path to the cached file.

    Examples:
        Cache the file at ``/srv/data.csv`` as ``data.csv``.

        >>> cache_file('/srv/data.csv', 'data.csv')
    """
    os.makedirs(CACHE_DIR, exist_ok=True)
    local_path = os.path.join(CACHE_DIR, local_name)
    if not os.path.isfile(local_path):
        shutil.copyfile(srv_path, local_path)
    return local_path


def copy_file(src: str, dest: str, checksum: Optional[bool] = False):
    """Copies a file from the source path to the destination path.

    Args:
        src: The source path.
        dest: The destination path.
        checksum: Whether to validate the checksum of the file.

    Raises:
        IOError: If the checksum is invalid.
    """
    shutil.copyfile(src, dest)
    if checksum:
        if md5sum(src) != md5sum(dest):
            raise IOError(f'Hash mismatch: {src} != {dest}')


def move_file(src: str, dest: str, checksum: Optional[bool] = False):
    """Moves a file from the source path to the destination path.

    Args:
        src: The source path.
        dest: The destination path.
        checksum: Whether to validate the checksum of the file.

    Raises:
        IOError: If the checksum is invalid.
    """
    copy_file(src, dest, checksum)
    os.remove(src)



def canonical_gid(gid: str) -> str:
    """Get canonical form of NCBI genome accession.

    Args:
        gid: The NCBI genome accession.

    Example:
        >>> canonical_gid('GCF_005435135.1_ASM543513v1_genomic')
        'G005435135'
    """

    if gid.startswith('U'):
        return gid

    gid = gid.replace('RS_', '').replace('GB_', '')
    gid = gid.replace('GCA_', 'G').replace('GCF_', 'G')
    if '.' in gid:
        gid = gid[0:gid.find('.')]

    return gid



import os
import shutil
import tempfile

import dendropy

from magna.config import MAGNA_DIR
from magna.util.io import download_file, md5sum


class GtdbTree:
    """The base class that all GTDB tree objects inherit."""

    __slots__ = ('source', 'path', 'md5', 'tree')

    def __init__(self, source: str, path: str, md5: str):
        #: The source URL.
        self.source: str = source
        #: The path to the downloaded file.
        self.path: str = path
        #: The expected MD5 checksum of the file.
        self.md5: str = md5
        if not os.path.isfile(self.path):
            self._download()
        #: The dendropy tree.
        self.tree: dendropy.Tree = self._read()

    def _read(self):
        return dendropy.Tree.get(path=self.path, schema='newick', preserve_underscores=True)

    def _download(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            # Download
            tmp_path = os.path.join(tmpdir, 'tree.tree')
            download_file(self.source, tmp_path)

            # Extract and validate
            if md5sum(tmp_path) != self.md5:
                raise ValueError('MD5 checksum mismatch')

            os.makedirs(os.path.dirname(self.path), exist_ok=True)
            shutil.copyfile(tmp_path, self.path)


class GtdbTreeR95Arc(GtdbTree):
    """The GTDB archaeal tree (release 95)."""
    source = 'https://data.gtdb.ecogenomic.org/releases/release95/95.0/ar122_r95.tree'
    path = os.path.join(MAGNA_DIR, 'dataset', 'gtdb', 'tree', 'ar122_r95.tree')
    md5 = '2f5e072b9095617e7b5cff09653f8bec'

    def __init__(self):
        super().__init__(self.source, self.path, self.md5)


class GtdbTreeR95Bac(GtdbTree):
    """The GTDB bacterial tree (release 95)."""
    source = 'https://data.gtdb.ecogenomic.org/releases/release95/95.0/bac120_r95.tree'
    path = os.path.join(MAGNA_DIR, 'dataset', 'gtdb', 'tree', 'bac120_r95.tree')
    md5 = 'c896d0eece01b281e09bd38534cd072e'

    def __init__(self):
        super().__init__(self.source, self.path, self.md5)


class GtdbTreeR202Arc(GtdbTree):
    """The GTDB archaeal tree (release 202)."""
    source = 'https://data.gtdb.ecogenomic.org/releases/release202/202.0/ar122_r202.tree'
    path = os.path.join(MAGNA_DIR, 'dataset', 'gtdb', 'tree', 'ar122_r202.tree')
    md5 = '5b2dd87b0836fd63a223a556eae2906d'

    def __init__(self):
        super().__init__(self.source, self.path, self.md5)


class GtdbTreeR202Bac(GtdbTree):
    """The GTDB bacterial tree (release 202)."""
    source = 'https://data.gtdb.ecogenomic.org/releases/release202/202.0/bac120_r202.tree'
    path = os.path.join(MAGNA_DIR, 'dataset', 'gtdb', 'tree', 'bac120_r202.tree')
    md5 = 'aebfc092ff6f2d81ef1226da6f1477c9'

    def __init__(self):
        super().__init__(self.source, self.path, self.md5)



import numpy as np
import pandas as pd

DIAMOND_DTYPE = {
    'query': object,
    'reference': object,
    'identity': np.float64,
    'length': np.uintc,
    'mismatches': np.uintc,
    'gap_openings': np.uintc,
    'query_start': np.uintc,
    'query_end': np.uintc,
    'target_start': np.uintc,
    'target_end': np.uintc,
    'e_value': np.float64,
    'bit_score': np.float64,
}


def read_diamond_output(path: str) -> pd.DataFrame:
    """Return a pandas DataFrame from a DIAMOND output file.

    Columns:
        * ``query`` - the accession of the sequence that was searched against the database, as specified in the input FASTA file after the > character until the first blank.
        * ``reference`` - the accession of the target database sequence that the query was aligned against
        * ``identity`` - the percentage of identical amino acid residues that were aligned against each other in the local alignment
        * ``length`` - the total length of the local alignment, which including matching and mismatching positions of query and subject, as well as gap positions in the query and subject.
        * ``mismatches`` - the number of non-identical amino acid residues aligned against each other.
        * ``gap_openings`` - the number of gap openings.
        * ``query_start`` - the starting coordinate of the local alignment in the query (1-based).
        * ``query_end`` - the ending coordinate of the local alignment in the query (1-based).
        * ``target_start`` - the starting coordinate of the local alignment in the subject (1-based).
        * ``target_end`` - the ending coordinate of the local alignment in the subject (1-based).
        * ``e_value`` - the expected value of the hit quantifies the number of alignments of similar or better quality that you expect to find searching this query against a database of random sequences the same size as the actual target database. This number is most useful for measuring the significance of a hit. By default, DIAMOND will report all alignments with e-value < 0.001, meaning that a hit of this quality will be found by chance on average once per 1,000 queries.
        * ``bit_score`` - the bit score is a scoring matrix independent measure of the (local) similarity of the two aligned sequences, with higher numbers meaning more similar. It is always >= 0 for local Smith Waterman alignments.

    Args:
        path: The path to the DIAMOND output file.
    """
    return pd.read_csv(path, sep='\t', header=None, names=DIAMOND_DTYPE.keys(),
                       dtype=DIAMOND_DTYPE)



import os
import tempfile

import pandas as pd

from magna.config import MAGNA_DIR
from magna.util.io import download_file, md5sum, untar


class GtdbMetadata:

    def __init__(self, source: str, path: str, md5: str):
        #: The source URL
        self.source: str = source
        #: The path to the metadata file.
        self.path: str = path
        #: The MD5 checksum of the downloaded file.
        self.md5: str = md5
        if not os.path.isfile(self.path):
            self._download()
        #: The metadata as a pandas DataFrame.
        self.df: pd.DataFrame = self._read()

    def _read(self):
        df = pd.read_feather(self.path)
        df.set_index('accession')
        return df

    @staticmethod
    def _read_tsv(path):
        return pd.read_csv(path, sep='\t', index_col=False, low_memory=False)

    def _download(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            # Download
            tmp_path_gz = os.path.join(tmpdir, 'metadata.tar.gz')
            download_file(self.source, tmp_path_gz)

            # Extract and validate
            untar(tmp_path_gz, tmpdir)
            tmp_path = os.path.join(tmpdir, os.path.basename(self.path).replace('.feather', '.tsv'))
            if md5sum(tmp_path) != self.md5:
                raise ValueError('MD5 checksum mismatch')

            df = self._read_tsv(tmp_path)
            os.makedirs(os.path.dirname(self.path), exist_ok=True)
            df.to_feather(path=self.path, compression='lz4')


# ----------------------------------------------------------------------------------------------------------------------

class GtdbMetadataR95Arc(GtdbMetadata):
    """The archaeal metadata (release 95)."""

    source = 'https://data.gtdb.ecogenomic.org/releases/release95/95.0/ar122_metadata_r95.tar.gz'
    path = os.path.join(MAGNA_DIR, 'dataset', 'gtdb', 'metadata', 'ar122_metadata_r95.feather')
    md5 = '110ad5daa2dbed2ee904b10c295da5dc'

    def __init__(self):
        super().__init__(self.source, self.path, self.md5)


class GtdbMetadataR95Bac(GtdbMetadata):
    """The bacterial metadata (release 95)."""

    source = 'https://data.gtdb.ecogenomic.org/releases/release95/95.0/bac120_metadata_r95.tar.gz'
    path = os.path.join(MAGNA_DIR, 'dataset', 'gtdb', 'metadata', 'bac120_metadata_r95.feather')
    md5 = '223ada02ffca4d1a2dda6edb9a164dcd'

    def __init__(self):
        super().__init__(self.source, self.path, self.md5)


class GtdbMetadataR95:
    """The combined archaeal and bacterial metadata (release 95)."""

    def __init__(self):
        #: The combined dataframe.
        self.df: pd.DataFrame = pd.concat([GtdbMetadataR95Arc().df, GtdbMetadataR95Bac().df])


# ----------------------------------------------------------------------------------------------------------------------

class GtdbMetadataR202Arc(GtdbMetadata):
    """The archaeal metadata (release 202)."""

    source = 'https://data.gtdb.ecogenomic.org/releases/release202/202.0/ar122_metadata_r202.tar.gz'
    path = os.path.join(MAGNA_DIR, 'dataset', 'gtdb', 'metadata', 'ar122_metadata_r202.feather')
    md5 = '0607728ae1f56bdb1a7cc24d238185c3'

    def __init__(self):
        super().__init__(self.source, self.path, self.md5)


class GtdbMetadataR202Bac(GtdbMetadata):
    """The bacterial metadata (release 202)."""
    source = 'https://data.gtdb.ecogenomic.org/releases/release202/202.0/bac120_metadata_r202.tar.gz'
    path = os.path.join(MAGNA_DIR, 'dataset', 'gtdb', 'metadata', 'bac120_metadata_r202.feather')
    md5 = '68fed11eb688982edb6f4669476c2a10'

    def __init__(self):
        super().__init__(self.source, self.path, self.md5)


class GtdbMetadataR202:
    """The combined archaeal and bacterial metadata (release 202)."""

    def __init__(self):
        #: The combined dataframe.
        self.df: pd.DataFrame = pd.concat([GtdbMetadataR202Arc().df, GtdbMetadataR202Bac().df])

# ----------------------------------------------------------------------------------------------------------------------



import os
import tempfile

import numpy as np
import pandas as pd

from magna.config import MAGNA_DIR
from magna.util.io import download_file


def read_contig_assignments_tsv(path: str) -> pd.DataFrame:
    """Read from the GUNC contig_assignments output file.

    Args:
        path: The path to the DIAMOND output file.
    """
    dtype = {
        'contig': object,
        'tax_level': object,
        'assignment': object,
        'count_of_genes_assigned': np.uintc
    }
    return pd.read_csv(path, sep='\t', dtype=dtype)


def gunc_max_css_scores_gtdb_r95() -> pd.DataFrame:
    """Return the max clade separation score (CSS) for the R95 GTDB."""
    path = os.path.join(MAGNA_DIR, 'gunc', 'gtdb_95.maxcss_level.feather')
    if not os.path.isfile(path):
        raise IOError(f'{path} does not exist.')
    return pd.read_feather(path)


def gunc_contig_assignment_gtdb_r95() -> pd.DataFrame:
    """Return the contig assignment for the R95 GTDB."""
    path = os.path.join(MAGNA_DIR, 'gunc', 'gtdb_95.contig_assignments.feather')
    if not os.path.isfile(path):
        raise IOError(f'{path} does not exist.')
    return pd.read_feather(path)


def gunc_all_levels_gtdb_r95() -> pd.DataFrame:
    """Return GUNC output at all levels for the R95 GTDB."""
    path = os.path.join(MAGNA_DIR, 'gunc', 'gtdb_95.all_levels.feather')
    if not os.path.isfile(path):
        raise IOError(f'{path} does not exist.')
    return pd.read_feather(path)


class GuncMaxCssScores:
    """Return the max clade separation score (CSS) for the R95 GTDB (progenes db)."""

    source = 'https://swifter.embl.de/~fullam/gunc/paper_supplementary_files/All_Datasets.GUNC.scores.maxCSS_level.tsv'
    path = os.path.join(MAGNA_DIR, 'dataset', 'gunc', 'All_Datasets.GUNC.scores.maxCSS_level.feather')
    md5 = 'dd91aa177b9112c361b9503e132f1c06'

    def __init__(self):
        if not os.path.isfile(self.path):
            self._download()
        #: The dataframe.
        self.df: pd.DataFrame = self._read()

    def _read(self):
        print('Note: Only the RefSeq and GenBank results are used.')
        return pd.read_feather(self.path)

    @staticmethod
    def _read_tsv(path):
        dtype = {
            'genome': np.object,
            'n_genes_called': np.uintc,
            'n_genes_mapped': np.uintc,
            'n_contigs': np.uintc,
            'taxonomic_level': np.object,
            'proportion_genes_retained_in_major_clades': np.float16,
            'genes_retained_index': np.float16,
            'clade_separation_score': np.float16,
            'contamination_portion': np.float16,
            'n_effective_surplus_clades': np.float16,
            'mean_hit_identity': np.float16,
            'reference_representation_score': np.float16,
            'pass.GUNC': np.object,
            'study': np.object,
            'CheckM_completeness': np.object,
            'CheckM_contamination': np.object,
        }
        rows = list()
        allowed_studies = frozenset({'GenBank', 'RefSeq'})
        with open(path, 'r') as f:
            header = {k: i for i, k in enumerate(
                f.readline().strip().split('\t'))}
            study_idx = header['study']
            for line in f.readlines():
                cols = line.strip().split('\t')
                if cols[study_idx] in allowed_studies:
                    rows.append(cols)
        return pd.DataFrame(rows, columns=dtype)

    def _download(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            # Download the file
            tmp_path = os.path.join(tmpdir, 'download.tsv')
            download_file(self.source, tmp_path, self.md5)

            df = self._read_tsv(tmp_path)
            os.makedirs(os.path.dirname(self.path), exist_ok=True)
            df.to_feather(path=self.path, compression='lz4')


class GuncAllScores:
    """Return GUNC output at all levels for the R95 GTDB (progenes db)."""

    source = 'https://swifter.embl.de/~fullam/gunc/paper_supplementary_files/All_Datasets.GUNC.scores.all_levels.specI2species.tsv'
    path = os.path.join(MAGNA_DIR, 'dataset', 'gunc', 'All_Datasets.GUNC.scores.all_levels.specI2species.feather')
    md5 = 'a54e3719221a42a5f96b267412827d27'

    def __init__(self):
        if not os.path.isfile(self.path):
            self._download()
        #: The dataframe.
        self.df: pd.DataFrame = self._read()

    def _read(self):
        print('Note: Only the RefSeq and GenBank results are used.')
        return pd.read_feather(self.path)

    @staticmethod
    def _read_tsv(path):
        dtype = {
            'genome': np.object,
            'n_genes_called': np.uintc,
            'n_genes_mapped': np.uintc,
            'n_contigs': np.uintc,
            'taxonomic_level': np.object,
            'proportion_genes_retained_in_major_clades': np.float16,
            'genes_retained_index': np.float16,
            'clade_separation_score': np.float16,
            'contamination_portion': np.float16,
            'n_effective_surplus_clades': np.float16,
            'mean_hit_identity': np.float16,
            'reference_representation_score': np.float16,
            'pass.GUNC': np.object,
            'study': np.object,
        }
        rows = list()
        allowed_studies = frozenset({'GenBank', 'RefSeq'})
        with open(path, 'r') as f:
            header = {k: i for i, k in enumerate(
                f.readline().strip().split('\t'))}
            study_idx = header['study']
            for line in f.readlines():
                cols = line.strip().split('\t')
                if cols[study_idx] in allowed_studies:
                    rows.append(cols)
        return pd.DataFrame(rows, columns=dtype)

    def _download(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            # Download the file
            tmp_path = os.path.join(tmpdir, 'download.tsv')
            download_file(self.source, tmp_path, self.md5)

            df = self._read_tsv(tmp_path)
            os.makedirs(os.path.dirname(self.path), exist_ok=True)
            df.to_feather(path=self.path, compression='lz4')



import os
from typing import Dict, Tuple

from Bio import SeqIO
from Bio.SeqRecord import SeqRecord

from magna.gtdb.enums import GtdbRelease
from magna.util.io import cache_file


class Genome:
    """A wrapper to a GTDB genome."""

    def __init__(self, accession: str, root: str):
        #: The short accession of the genome.
        self.accession: str = accession
        #: The root directory where this genome is stored.
        self.root: str = root

        # Generate paths
        base = os.path.basename(self.root)
        #: The path to the CDS file.
        self.cds_path: str = os.path.join(self.root, f'{base}_cds_from_genomic.fna')
        #: The path to the FNA file.
        self.fna_path: str = os.path.join(self.root, f'{base}_genomic.fna')

    def __repr__(self):
        return str(self.accession)

    def cds_seqio(self) -> Tuple[SeqRecord, ...]:
        """Read and return the CDS file as a SeqIO object."""
        with open(self.cds_path, 'r') as f:
            out = tuple(SeqIO.parse(f, 'fasta'))
        return out

    def fna_seqio(self) -> Tuple[SeqRecord, ...]:
        """Read and return the FNA file as a SeqIO object."""
        with open(self.fna_path, 'r') as f:
            out = tuple(SeqIO.parse(f, 'fasta'))
        return out


class GenomeDirs:
    """An interface to the :obj:`GtdbRelease` accession to :obj:`Genome` mapping."""

    def __init__(self, release: GtdbRelease):
        """Initialise the GenomeDirs class for a given release.

        Args:
            release: The release of GTDB to use.
        """
        self.release = release

        # Create the paths
        srv_path = f'/srv/db/gtdb/genomes/ncbi/release{release.value}/genome_dirs.tsv'
        cache_path = cache_file(srv_path, f'genome_dirs_{self.release.value}.tsv')

        # Read the data
        self._data = self.read(cache_path)

    @staticmethod
    def read(path: str) -> Dict[str, str]:
        out = dict()
        with open(path, 'r') as f:
            for line in f:
                short, root, canonical = line.strip().split('\t')
                out[short] = root
        return out

    def get(self, accession: str) -> Genome:
        """Return the :obj:`Genome` for the given accession.

        Args:
            accession: The short accession of the genome (e.g. ``GCA_123456789.1``).
        """
        return Genome(accession=accession, root=self._data[accession])
